package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewCatalog
import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.rest.files.FILE_CONTROLLER_PATH
import dev.fyloz.colorrecipesexplorer.service.MaterialService
import org.springframework.http.MediaType
import org.springframework.http.ResponseEntity
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import org.springframework.web.multipart.MultipartFile
import java.net.URI
import java.net.URLEncoder
import java.nio.charset.StandardCharsets
import javax.validation.Valid

private const val MATERIAL_CONTROLLER_PATH = "api/material"

@RestController
@RequestMapping(MATERIAL_CONTROLLER_PATH)
@PreAuthorizeViewCatalog
class MaterialController(
    private val materialService: MaterialService,
    private val creProperties: CreProperties
) {
    @GetMapping
    fun getAll() =
        ok(materialService.getAll())

    @GetMapping("notmixtype")
    fun getAllNotMixType() =
        ok(materialService.getAllNotMixType())

    @GetMapping("{id}")
    fun getById(@PathVariable id: Long) =
        ok(materialService.getById(id))

    @PostMapping(consumes = [MediaType.MULTIPART_FORM_DATA_VALUE])
    @PreAuthorize("hasAuthority('EDIT_MATERIALS')")
    fun save(@Valid material: MaterialSaveDto, simdutFile: MultipartFile?) =
        created {
            materialService.save(
                materialSaveDto(
                    name = material.name,
                    inventoryQuantity = material.inventoryQuantity,
                    materialTypeId = material.materialTypeId,
                    simdutFile = simdutFile
                )
            )
        }

    @PutMapping(consumes = [MediaType.MULTIPART_FORM_DATA_VALUE])
    @PreAuthorize("hasAuthority('EDIT_MATERIALS')")
    fun update(@Valid material: MaterialUpdateDto, simdutFile: MultipartFile?) =
        noContent {
            materialService.update(
                materialUpdateDto(
                    id = material.id,
                    name = material.name,
                    inventoryQuantity = material.inventoryQuantity,
                    materialTypeId = material.materialTypeId,
                    simdutFile = simdutFile
                )
            )
        }

    @DeleteMapping("{id}")
    @PreAuthorize("hasAuthority('REMOVE_MATERIALS')")
    fun deleteById(@PathVariable id: Long) =
        noContent {
            materialService.deleteById(id)
        }

    @GetMapping("mix/create/{recipeId}")
    fun getAllForMixCreation(@PathVariable recipeId: Long) =
        ok(materialService.getAllForMixCreation(recipeId))

    @GetMapping("mix/update/{mixId}")
    fun getAllForMixUpdate(@PathVariable mixId: Long) =
        ok(materialService.getAllForMixUpdate(mixId))

    private fun ok(material: Material) =
        ok(material.toOutput())

    private fun ok(materials: Collection<Material>) =
        ok(materials.map { it.toOutput() })

    private fun created(producer: () -> Material): ResponseEntity<MaterialOutputDto> = with(producer().toOutput()) {
        ResponseEntity
            .created(URI.create("$MATERIAL_CONTROLLER_PATH/${this.id}"))
            .body(this)
    }

    private fun Material.toOutput() = materialOutputDto(
        this,
        if (materialService.hasSimdut(this)) this.simdutUrl() else null
    )

    private fun Material.simdutUrl() =
        "${creProperties.deploymentUrl}$FILE_CONTROLLER_PATH?path=${
            URLEncoder.encode(
                this.simdutFilePath,
                StandardCharsets.UTF_8
            )
        }"
}

