package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.validation.or
import dev.fyloz.colorrecipesexplorer.repository.RecipeRepository
import dev.fyloz.colorrecipesexplorer.service.files.FileService
import dev.fyloz.colorrecipesexplorer.service.utils.setAll
import org.springframework.context.annotation.Lazy
import org.springframework.stereotype.Service
import org.springframework.web.multipart.MultipartFile
import java.io.File
import java.nio.file.NoSuchFileException
import javax.transaction.Transactional

interface RecipeService : ExternalModelService<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeRepository> {
    /** Checks if one or more recipes have the given [company]. */
    fun existsByCompany(company: Company): Boolean

    /** Gets all recipes with the given [company]. */
    fun getAllByCompany(company: Company): Collection<Recipe>

    /** Updates the public data of a recipe with the given [publicDataDto]. */
    fun updatePublicData(publicDataDto: RecipePublicDataDto)

    /** Adds the given [mix] to the given [recipe]. */
    fun addMix(recipe: Recipe, mix: Mix): Recipe

    /** Removes the given [mix] from its recipe. */
    fun removeMix(mix: Mix): Recipe
}

@Service
class RecipeServiceImpl(
    recipeRepository: RecipeRepository,
    val companyService: CompanyService,
    val mixService: MixService,
    val recipeStepService: RecipeStepService,
    @Lazy val groupService: EmployeeGroupService
) :
    AbstractExternalModelService<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeRepository>(recipeRepository),
    RecipeService {
    override fun idNotFoundException(id: Long) = recipeIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = recipeIdAlreadyExistsException(id)

    override fun existsByCompany(company: Company): Boolean = repository.existsByCompany(company)
    override fun getAllByCompany(company: Company): Collection<Recipe> = repository.findAllByCompany(company)

    override fun save(entity: RecipeSaveDto): Recipe {
        // TODO checks if name is unique in the scope of the [company]
        return save(with(entity) {
            recipe(
                name = name,
                description = description,
                color = color,
                gloss = gloss,
                sample = sample,
                approbationDate = approbationDate,
                remark = remark ?: "",
                company = companyService.getById(companyId)
            )
        })
    }

    @Transactional
    override fun update(entity: RecipeUpdateDto): Recipe {
        val persistedRecipe by lazy { getById(entity.id) }

        return update(with(entity) {
            recipe(
                id = id,
                name = name or persistedRecipe.name,
                description = description or persistedRecipe.description,
                color = color or persistedRecipe.color,
                gloss = gloss ?: persistedRecipe.gloss,
                sample = sample ?: persistedRecipe.sample,
                approbationDate = approbationDate ?: persistedRecipe.approbationDate,
                remark = remark or persistedRecipe.remark,
                company = persistedRecipe.company,
                mixes = persistedRecipe.mixes,
                groupsInformation = updateGroupsInformation(persistedRecipe, entity)
            )
        })
    }

    private fun updateGroupsInformation(recipe: Recipe, updateDto: RecipeUpdateDto): Set<RecipeGroupInformation> {
        val steps = updateDto.steps ?: return recipe.groupsInformation

        val updatedGroupsInformation = mutableSetOf<RecipeGroupInformation>()
        steps.forEach {
            with(recipe.groupInformationForGroup(it.groupId)) {
                // Set steps for the existing RecipeGroupInformation or create a new one
                val updatedGroupInformation = this?.apply {
                    if (this.steps != null) {
                        this.steps!!.setAll(it.steps)
                    } else {
                        this.steps = it.steps.toMutableSet()
                    }
                } ?: recipeGroupInformation(
                    group = groupService.getById(it.groupId),
                    steps = it.steps.toMutableSet()
                )

                updatedGroupsInformation.add(updatedGroupInformation)
                recipeStepService.validateGroupInformationSteps(updatedGroupInformation)
            }
        }

        return updatedGroupsInformation
    }

    @Transactional
    override fun updatePublicData(publicDataDto: RecipePublicDataDto) {
        if (publicDataDto.notes != null) {
            val recipe = getById(publicDataDto.recipeId)

            fun noteForGroup(group: EmployeeGroup) =
                publicDataDto.notes.firstOrNull { it.groupId == group.id }?.content

            // Notes
            recipe.groupsInformation.map {
                val updatedNote = noteForGroup(it.group)
                it.apply {
                    note = updatedNote
                }
            }

            update(recipe)
        }

        if (publicDataDto.mixesLocation != null) {
            mixService.updateLocations(publicDataDto.mixesLocation)
        }
    }

    override fun addMix(recipe: Recipe, mix: Mix) =
        update(recipe.apply { mixes.add(mix) })

    override fun removeMix(mix: Mix): Recipe =
        update(mix.recipe.apply { mixes.remove(mix) })
}

const val RECIPE_IMAGES_DIRECTORY = "images/recipe"

interface RecipeImageService {
    // TOOD change return type to ByteArrayResource
    fun getByIdForRecipe(id: Long, recipeId: Long): ByteArray

    /** Gets the identifier of every images associated to the recipe with the given [recipeId]. */
    fun getAllIdsForRecipe(recipeId: Long): Collection<Long>

    /** Saves the given [image] and associate it to the recipe with the given [recipeId]. Returns the identifier of the saved image. */
    fun save(image: MultipartFile, recipeId: Long): Long

    /** Deletes the image with the given [recipeId] and [id]. */
    fun delete(id: Long, recipeId: Long)
}

@Service
class RecipeImageServiceImpl(val recipeService: RecipeService, val fileService: FileService) : RecipeImageService {
    override fun getByIdForRecipe(id: Long, recipeId: Long): ByteArray =
        try {
            fileService.read(getPath(id, recipeId)).byteArray
        } catch (ex: NoSuchFileException) {
            throw RecipeImageNotFoundException(id, recipeService.getById(recipeId))
        }

    override fun getAllIdsForRecipe(recipeId: Long): Collection<Long> {
        val recipe = recipeService.getById(recipeId)
        val recipeDirectory = getRecipeDirectory(recipe.id!!)
        if (!recipeDirectory.exists() || !recipeDirectory.isDirectory) {
            return listOf()
        }
        return recipeDirectory.listFiles()!!   // Should never be null because we check if recipeDirectory is a directory and exists before
            .filterNotNull()
            .map { it.name.toLong() }
    }

    override fun save(image: MultipartFile, recipeId: Long): Long {
        /** Gets the next id available for a new image for the recipe with the given [recipeId]. */
        fun getNextAvailableId(): Long =
            with(getAllIdsForRecipe(recipeId)) {
                if (isEmpty())
                    0
                else
                    maxOrNull()!! + 1L  // maxOrNull() cannot return null because existingIds cannot be empty at this point
            }

        val nextAvailableId = getNextAvailableId()
        fileService.write(image, getPath(nextAvailableId, recipeId), true)
        return nextAvailableId
    }

    override fun delete(id: Long, recipeId: Long) =
        fileService.delete(getPath(id, recipeId))

    /** Gets the images directory of the recipe with the given [recipeId]. */
    fun getRecipeDirectory(recipeId: Long) = File("$RECIPE_IMAGES_DIRECTORY/$recipeId")

    /** Gets the file of the image with the given [recipeId] and [id]. */
    fun getPath(id: Long, recipeId: Long): String = "$RECIPE_IMAGES_DIRECTORY/$recipeId/$id"
}
